
# sliderule: The small web toolkit for Go

sliderule is the toolkit for working with the small web in Go.

```
 _____________________________________________________________
|  |"!"|"!"|"!"|"!"|"!"|"!"|"!"|"!"|"!"|"!"|"!"|"!"|"!"|"!"|  |
|  0   1   2   3   4   5   6   7   8   9  10  11  12  13   4  |
|---------------------------=====+=====-----------------------|
|___:981,:,!,,2::!:3n:4:5:6|7891:|:!:!2|:!.3:!:4!5:67891!:!___|mmm
    |   |891"!"|""2""|"3n|4|5"6|7891"!"|"|2""!"3"|"4|5"67891"!|   |
    |                      |     I     |                          |
 ___|___9:1:1c2:3:4:5:67892|::!::|3n:C:|4::!:5:::6::7::8:9:1:1c___|
|   9!1"1"2"3"4"5"67892"|"||""3n"|""4""|"5"!"6""7""8"9"1"1    |===
|__________________________|_____|__apx|______________________|
                            =====+=====
```

Think of it as a net/http for small web protocols. You still have to write your server, but you can focus on the logic you want to implement knowing the protocol is already dealt with. It's been said of gemini that you can write your server in a day. Now you can write it in well under an hour.

## The slide rule

sliderule is named after the iconic instrument used to perform approximate mathematical calculations before the advent of handheld electronic calculators. It was the predominant tool that identified an engineer during project Gemini, a sort of developer's tool of it's time.

## The "sliderule" package

sliderule is carefully structured as composable building blocks. The top-level package defines the framework in which servers and clients can be built.

* a request type
* a response type
* a "Server" interface type
* a "Handler" abstraction
* a "Middleware" abstraction
* a "Client" which handles all the protocols known to sliderule
* some useful Handler wrappers: a router, request filtering, falling through a list of handlers

## Protocols

The packages sliderule/gemini, sliderule/gopher, sliderule/finger, and sliderule/spartan provide concrete implementations specific to those protocols.

* I/O (parsing, formatting) request and responses
* constructors for the various kinds of protocol responses
* helpers for building a protocol-suitable TLS config
* Client implementations
* Servers which can run your Handlers.

The primary text formats for those protocols have higher-level support provided in sub-packages:

* sliderule/gemini/gemtext supports parsing gemtext and getting direct programmatic access to its AST. Deeper sub-packages provide converters to other formats (markdown and HTML) with overridable templates.
* sliderule/gopher/gophermap similarly parses the gophermap format and provides access to its AST.

## Logging

sliderule borrows the logging interface from go-kit.

=> [The logger interface from go-kit/log.](https://pkg.go.dev/github.com/go-kit/log#Logger)

The sliderule/logging package provides everything you need to get a good basic start to producing helpful logs.

* A request-logging middleware with common diagnostics (time, duration, url, status codes, response body lengths)
* A simple constructor of useful default loggers at various levels. They output colorful logfmt lines to stdout.

## Routing

The router in the sliderule package supports slash-delimited path pattern strings. In the segments of these patterns:

* A "/:wildcard/" segment matches anything in that position, and captures the value as a route parameter. Or if the paramter name is omitted like "/:/", it matches anything in a single segment without capturing a paramter.
* A "/*remainder" segment is only allowed at the end and matches the rest of the path, capturing it into the paramter name. Or again, omitting a parameter name like "/*" simple matches any path suffix.
* Any other segment in the pattern must match the corresponding segment of a request exactly.

Router also supports maintaining a list of middlewares at the router level, mounting sub-routers under a pattern, looking up the matching handler for any request, and of course acting as a Handler itself.

## sliderule/contrib/*

This is where useful building blocks themselves start to come in. Sub-packages of contrib include Handler and Middleware implementations which accomplish the things your servers actually need to do.

The sub-packages include:

* fs has handlers that make file servers possible: serve files, build directory listings, etc
* cgi includes handlers which can execute CGI programs
* sharedhost which provides means of handling /~username URLs
* tlsauth contains middlewares and bool functions for authenticating against TLS client certificates
* ...with more to come

## Get it

### Using sliderule in your project

To add it to your own go project:

```
$ go get tildegit.org/tjp/sliderule
```

### Straight to the code please

=> [The code is hosted here on tildegit.](https://tildegit.org/tjp/sliderule)

=> [The generated documentation is on the go package index.](https://pkg.go.dev/tildegit.org/tjp/sliderule)

### Verify releases

Releases are signed with minisign. The signature file is included in the release downloads page, and the public key is RWSzQywJwHgjSMD0y0RXwXAGpapcMJplwbCVYQqabhAJ+NAnKAeh98Vb - this is also referenced on tjp's gemini home page.

=> [tjp's home page, which also mentions the public key used for signing sliderule releases.](gemini://gemini.ctrl-c.club/~tjp)

## Previously

This project used to be called "gus", and the original project's old issues can be found at:

=> [https://tildegit.org/tjp/gus](https://tildegit.org/tjp/gus)

## Contribute

There's lots still to do, and contributions are very welcome!

=> [submit an issue or pull request on the tildegit repository,](https://tildegit.org/tjp/sliderule)

=> [send me an email directly,](mailto:tjp@ctrl-c.club)

or poke me on IRC: I'm @tjp on irc.tilde.chat where you'll find me in #gemini
